#property description "Linear Regression"
#property description "1 classic & 3 alternative methods"
#property copyright   "ds2"
#property version     "1.0"
//+------------------------------------------------------------------+
#property indicator_chart_window
#property indicator_buffers 5
#property indicator_plots   1
#property indicator_type1   DRAW_LINE
#property indicator_color1  Cyan
//+------------------------------------------------------------------+
enum LR_Method 
  {
   LR_M_Classic = 0,  // 
   LR_M_Sum     = 1,  //  
   LR_M_Func    = 2,  // 
   LR_M_Approx  = 3   // 
  };

//  

input LR_Method LRMethod  = LR_M_Sum;  //  
input int       LRPeriod  = 20;        //   
input bool      ShowLog   = false;     //   
//+------------------------------------------------------------------+
//   -    
double ExtLRBuffer[];

//     
double ExtBufSx[], ExtBufSy[], ExtBufSxx[], ExtBufSxy[];

//    
int h_SMA, h_LWMA;

//    
uint TicksStart, TicksEnd;
//+------------------------------------------------------------------+
void OnInit()
  {
   TicksStart=GetTickCount();   
  
   SetIndexBuffer(0, ExtLRBuffer, INDICATOR_DATA);
   PlotIndexSetInteger(0, PLOT_DRAW_BEGIN, LRPeriod-1);
   
   SetIndexBuffer(1, ExtBufSx,  INDICATOR_CALCULATIONS);
   SetIndexBuffer(2, ExtBufSy,  INDICATOR_CALCULATIONS);
   SetIndexBuffer(3, ExtBufSxx, INDICATOR_CALCULATIONS);
   SetIndexBuffer(4, ExtBufSxy, INDICATOR_CALCULATIONS);

   IndicatorSetString (INDICATOR_SHORTNAME,"Linear Regression");
   IndicatorSetInteger(INDICATOR_DIGITS,_Digits+1);
   
   //     
   if (LRMethod == LR_M_Func) {
      h_SMA  = iMA(NULL, 0, LRPeriod, 0, MODE_SMA,  PRICE_CLOSE);
      h_LWMA = iMA(NULL, 0, LRPeriod, 0, MODE_LWMA, PRICE_CLOSE);   
   }

   //     
   if (ShowLog)
     Print("OnInit().   ", TicksStart, ",   ", GetTickCount());
  }
//+------------------------------------------------------------------+
int OnCalculate(const int rates_total,
                const int prev_calculated,
                const int begin,
                const double &price[])
  {
   if (ShowLog & ((rates_total - prev_calculated)>2))
     Print("OnCalculate().   ", GetTickCount());   
   
   if (rates_total < LRPeriod)
      return(0);

   int limit = prev_calculated ? prev_calculated-1 : LRPeriod-1;
 
   bool IsExistsCalcBars = prev_calculated > 0;
   
   //    
   for (int bar = limit; bar < rates_total; bar++)
     {
      double lrvalue = 0; //      
      
      //===============================
      //  
      //   
      //  
      //===============================
      
      double Sx=0, Sy=0, Sxy=0, Sxx=0;
      
      if ( (LRMethod == LR_M_Classic) || ((LRMethod == LR_M_Sum)&&(!IsExistsCalcBars)) )
        //----------------------------------//
        //   
        //----------------------------------//      
        {
            //   -
            Sx  = 0;
            Sy  = 0;
            Sxx = 0;
            Sxy = 0;
            for (int x = 1; x <= LRPeriod; x++)
              {
               double y = price[bar-LRPeriod+x];
               Sx  += x;
               Sy  += y;
               Sxx += x*x;
               Sxy += x*y;
              }

            //  
            double a = (LRPeriod * Sxy - Sx * Sy) / (LRPeriod * Sxx - Sx * Sx);
            double b = (Sy - a * Sx) / LRPeriod;

            lrvalue = a*LRPeriod + b;
            
            //    LR_M_Sum,  ,   
            IsExistsCalcBars = true;
        }
        //----------------------------------//
      else if (LRMethod == LR_M_Sum)
        //----------------------------------//
        //   
        //----------------------------------//
        {
            // (      )        
        
            //  
            int prevbar = bar-1;
            
            //---      
            //        
            
            Sx  = ExtBufSx [prevbar]; 
            
            //   ,  
            Sy  = ExtBufSy [prevbar] - price[bar-LRPeriod] + price[bar]; 
            
            Sxx = ExtBufSxx[prevbar];
            
            //      ,      
            Sxy = ExtBufSxy[prevbar] - ExtBufSy[prevbar] + price[bar]*LRPeriod;
            
            //---

            //   (  ,    )
            double a = (LRPeriod * Sxy - Sx * Sy) / (LRPeriod * Sxx - Sx * Sx);
            double b = (Sy - a * Sx) / LRPeriod;

            lrvalue = a*LRPeriod + b;            
        }
        //----------------------------------//
      else if (LRMethod == LR_M_Func)
        //----------------------------------//
        //  
        //----------------------------------//      
        {
            double SMA [1];
            double LWMA[1];
            CopyBuffer(h_SMA,  0, rates_total-bar, 1, SMA);
            CopyBuffer(h_LWMA, 0, rates_total-bar, 1, LWMA);
            
            lrvalue = 3*LWMA[0] - 2*SMA[0];
        }
        //----------------------------------//
      else if (LRMethod == LR_M_Approx)
        //----------------------------------//
        //  
        //----------------------------------//
        {
           //  
           int HalfPeriod = (int) MathRound(LRPeriod/2);
           
           //    
           double s1 = 0;
           for (int i = 0; i < HalfPeriod; i++)
              s1 += price[bar-i];
           s1 /= HalfPeriod;
              
           //    
           double s2 = 0;
           for (int i = HalfPeriod; i < LRPeriod; i++)
              s2 += price[bar-i];
           s2 /= (LRPeriod-HalfPeriod);
           
           //     
           double k = (s1-s2)/(LRPeriod/2);
           
           //     
           lrvalue = s1 + k * (HalfPeriod-1)/2;
        }
        //----------------------------------//


      //      
      if (LRMethod == LR_M_Sum)
        {
         ExtBufSx  [bar] = Sx; 
         ExtBufSy  [bar] = Sy;
         ExtBufSxx [bar] = Sxx;
         ExtBufSxy [bar] = Sxy;
        }
        
      //   
      ExtLRBuffer[bar] = lrvalue;
     }

   //     
   TicksEnd = GetTickCount();
   if (ShowLog & ((rates_total - prev_calculated)>2))
     {
      Print("OnCalculate().   ", TicksEnd);
      Print(" : ", rates_total-limit,
            ".   (  OnInit), : ", TicksEnd-TicksStart);
     }

   return(rates_total);
  }
//+------------------------------------------------------------------+